/*
 * Access a SPI server to update a display panel backlight.
 *
 * This server is specific to the Ilitek ili8960 backlight controller, although
 * the Giantplus GPM940B0 panel datasheet also describes this controller
 * interface.
 *
 * Copyright (C) 2018, 2020 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/env>
#include <l4/sys/capability>
#include <l4/sys/err.h>

#include <stdio.h>

#include <l4/devices/spi-client.h>

#include <ipc/server.h>
#include "backlight_object_server.h"

/* SPI access abstractions. */

static L4::Cap<Spi_device_interface> spi_device;



/* Backlight device. */

class server_BacklightObject : public BacklightObject
{
  int _min = 55, _max = 90, _start = 70;

  void set_duty(int level)
  {
    level = level < _min ? _min : (level > _max ? _max : level);

    /* PWM_DUTY = R05h<5:3> = 5% increments from min to max */

    int duty = ((level - _min) / 5) << 3;

    spi_device->send(16, 0x0516 | duty); /* R05h: GRB=0 (reset); PWM_DUTY=duty; SHDB2=1, SHDB1=1 (power-related); STB=0 (standby) */
    spi_device->send(16, 0x0546 | duty); /* R05h: GRB=1 (normal operation); ... */
    spi_device->send(16, 0x078d);        /* R07h: HBLK=141 (horizontal blanking period from start of hsync pulse to data start) */
    spi_device->send(16, 0x1301);        /* R13h: IN_SEL=1 (alignment mode) */
    spi_device->send(16, 0x0547 | duty); /* R05h: ...; STB=1 (not standby) */
  }

public:
  /* Disable the backlight. */

  long disable()
  {
    spi_device->send(16, 0x0546); /* R05h: GRB=1 (normal operation); SHDB2=1, SHDB1=1 (power-related); STB=0 (standby) */
    return L4_EOK;
  }

  /* Enable the backlight. */

  long enable()
  {
    set_duty(_start);
    return L4_EOK;
  }

  /* Use the SPI device to update the brightness level. */

  long set_brightness(int level)
  {
    level = level < _min ? _min : (level > _max ? _max : level);
    spi_device->send(16, 0x0300 | level); /* R03h: brightness */
    return L4_EOK;
  }
};



int main(void)
{
  /* Obtain a reference to the SPI device. */

  spi_device = L4Re::Env::env()->get_cap<Spi_device_interface>("spi");
  if (!spi_device.is_valid()) return 1;

  /* Initialise and register a new server object. */

  server_BacklightObject obj;
  l4_cap_idx_t server;

  if (ipc_server_bind("backlight", (l4_umword_t) &obj, &server)) return 1;

  /* Enter the IPC server loop. */

  ipc_server_loop(BacklightObject_expected_items, &obj,
                  (ipc_server_handler_type) handle_BacklightObject);
  return 0;
}
